## Message Types
SVC_BAD=0
SVC_NOP=1
SVC_GAMESTATE=2
SVC_CONFIGSTRING=3
SVC_BASELINE=4
SVC_SERVERCOMMAND=5
SVC_DOWNLOAD=6
SVC_SNAPSHOT=7
SVC_EOF=8

## Gametypes
GT_FFA=0                         # free for all
GT_TOURNAMENT=1          # one on one tournament
GT_SINGLE_PLAYER=2       # single player ffa
GT_TEAM=3               # team deathmatch
GT_CTF=4                 # capture the flag
GT_1FCTF=5
GT_OBELISK=6
GT_HARVESTER=7

GT_STRING_MAP={GT_FFA: 'FFA',
               GT_TOURNAMENT: '1v1 Tournament',
               GT_SINGLE_PLAYER: 'Single Player FFA',
               GT_TEAM: 'Team DM',
               GT_CTF: 'CTF',
               GT_1FCTF: '1FCTF',
               GT_OBELISK: 'Obelisk',
               GT_HARVESTER: 'Harvester'}

def gametype_to_string(i):
    i=int(i)
    return GT_STRING_MAP.get(i, None)

MAX_MODELS = 242 # Not the same as Q3A
MAX_SOUNDS = 255 # Not the same as Q3A
MAX_CLIENTS = 64 
MAX_LOCATIONS = 64
MAX_TEAMS = 2 # Made up to filter team names from observed QL demos

# Configstring Definitions from Q3A source headers with mods for QL
CS_SERVERINFO = 0
CS_SYSTEMINFO = 1
CS_MUSIC = 2
CS_MESSAGE = 3               # from the map worldspawn's message field
CS_MOTD = 4               # g_motd string for server message of the day
CS_WARMUP = 5               # server time when the match will be restarted
CS_SCORES1 = 6
CS_SCORES2 = 7
CS_GAME_VERSION = 20
CS_LEVEL_START_TIME = 21   # so the timer only shows the current level
CS_INTERMISSION = 22       # when 1, fraglimit/timelimit has been hit and intermission will start in a second or two
CS_FLAGSTATUS = 23         # string indicating flag status in CTF
CS_SHADERSTATE = 24
CS_BOTINFO = 25
CS_ITEMS = 27              # string of 0's and 1's that tell which items are present
CS_MODELS = 32
CS_SOUNDS = (CS_MODELS+MAX_MODELS)
CS_PLAYERS = (CS_SOUNDS+MAX_SOUNDS)
CS_LOCATIONS = (CS_PLAYERS+MAX_CLIENTS)
CS_PARTICLES  = (CS_LOCATIONS+MAX_LOCATIONS)
CS_MAX = (CS_PARTICLES+MAX_LOCATIONS)
CS_TEAM = 659

CS_STRING_MAP = {CS_SERVERINFO: 'server_info',
                 CS_SYSTEMINFO: 'system_info',
                 CS_MUSIC: 'music',
                 CS_MESSAGE: 'map_message',
                 CS_MOTD: 'server_motd',
                 CS_WARMUP: 'match_restart',
                 CS_SCORES1: 'scores1',
                 CS_SCORES2: 'scores2',
                 CS_GAME_VERSION: 'game_version',
                 CS_LEVEL_START_TIME: 'level_start_time',
                 CS_INTERMISSION: 'intermission',
                 CS_FLAGSTATUS: 'ctf_flag_status',
                 CS_SHADERSTATE: 'shader_state',
                 CS_BOTINFO: 'bot_info',
                 CS_ITEMS: 'items_present',
                 CS_MODELS: 'models',
                 CS_SOUNDS: 'sounds',
                 CS_PLAYERS: 'players',
                 CS_LOCATIONS: 'locations',
                 CS_PARTICLES: 'particles'}

userinfo_map={'c1': 'color1',
              'c2': 'color2',
              'tt': 'team_target',
              'hc': 'handicap',
              'c': 'country',
              'cn': 'clan',
              'so': 'spectator_only',
              't': 'team',
              'xcn': 'extended_clan',
              'tl': 'team_leader'}



#!/bin/env python

## qldemo2json.py
## Shawn Nock, 2014

import argparse
import datetime
import json
import re
import os
import sys
import time

from qldemo import QLDemo, gametype_to_string
from qldemo.constants import userinfo_map, GT_TEAM

## Configuration ##
playerinfo_override = {'n': 'name',    ## The userInfo_t summary in
                       'cn': 'clan',   ## CS_PLAYER[MAX_PLAYERS] has short
                       'xcn': 'xclan', ## hard to decipher names. Any map
                       'w': 'wins',    ## here will override the name 
                       'l': 'losses',
                       't': 'team'}

### How should we extract and format a timestamp from the filename
timestamp_re = re.compile('([0-9]{4})_([0-9]{2})_([0-9]{2})-([0-9]{2})_([0-9]{2})_([0-9]{2})')
def timestamp(filename):
    hit = timestamp_re.search(filename)
    if not hit:
        return time.ctime(os.path.getctime(filename))
    return datetime.datetime.strptime(
        hit.group(0), 
        "%Y_%m_%d-%H_%M_%S").ctime()

### How should we extract/format a POV playername from a filename
pov_res = [re.compile('-([a-zA-Z0-9]*)\(POV\)-'),
           re.compile('-([a-zA-Z0-9]*)-'),
           re.compile('\[([a-zA-Z0-9]*)\]-')]
color_tag_re = re.compile('\^[0-9]')
def pov(d, filename):
    for regex in pov_res:
        hit = regex.search(filename)
        if not hit:
            continue
        for i in range(1,regex.groups+1):
            for player in d.gamestate['players']:
                if hit.group(i).find(color_tag_re.sub('', player['name'])) > -1:
                    return player['name']
    return None

def main():
    parser = argparse.ArgumentParser(
        description='Summarize, in JSON a QuakeLive Demo File (dm_73)')
    parser.add_argument('file',
                        help='path of the dm_73 file to summarize')
    args = parser.parse_args()

    d = QLDemo(args.file)
    gamestate=None
    for packet in d:
        if packet['type'] == 'gamestate':
            gamestate=packet
            break
    if not gamestate:
        return 1
    ## Munge playerInfo to conform to ColonelPanic's Needs
    for player in gamestate['players']:
        for key, value in player.iteritems():
            new_name=dict(userinfo_map.items()+playerinfo_override.items()).get(key, None)
            if new_name:
                player[new_name]=player[key]
                del(player[key])
        player['score']=""
        if int(gamestate['config']['server_info']['g_gametype']) >= GT_TEAM:
            player['team']=[team['name'] for team in d.gamestate['teams'] if team['id']==int(player['team'])][0]
    
    output = {'filename': args.file,
              'gametype': gametype_to_string(
                  gamestate['config']['server_info']['g_gametype']),
              'players': gamestate['players'],
              'size': os.stat(args.file).st_size,
              'by': pov(d, args.file),
              'timestamp': timestamp(args.file),
              'duration': None,
              'victor': None}
    
    if int(gamestate['config']['server_info']['g_gametype']) >= GT_TEAM:
        output['teams']=d.gamestate['teams']
        for team in output['teams']:
            team['score']=""
    
    json.dump(output, 
              sys.stdout, 
              ensure_ascii=False,
              indent=2)
    return 0

if __name__ == '__main__':
    main()





#!/bin/env python

## ql-demo-parse.py
## Shawn Nock, 2014

# Standard Lib
import json
import os
import re
import struct

# C-Extension wrapping Q3A Huffman Routines
import huffman

# Constants and enum maps
from qldemo.constants import *

# Classes

class PlayerState:
    pass

class ServerCommand:
    seq = None
    cmd = None
    string = None

    def __init__(self, seq, string):
        self.seq = seq
        self.cmd = string.split()[0]
        self.string = ' '.join(string.split()[1:])

 
class QLDemo:
    gamestate={
        'type': 'gamestate',
        'configstrings': {},
        'config': {},
        'players': [],
        'spectators': [],
        'teams': []}
    def __init__(self, filename):
        huffman.init()
        huffman.open(filename)

    def __iter__(self):
        while True:
            seq=huffman.readrawlong()
            length=huffman.readrawlong()
            if seq == -1 or length == -1:
                break
            huffman.fill(length)
            ack = huffman.readlong()
            cmd = huffman.readbyte()
            if cmd == SVC_GAMESTATE: 
                r = self.parse_gamestate()
            elif cmd == SVC_SERVERCOMMAND: 
                r = self.parse_servercommand()
            elif cmd == SVC_SNAPSHOT:
                continue
                r = self.parse_playerstate()
            yield r

    def parse_gamestate(self):
        ack=huffman.readlong()
        while True:
            cmd = huffman.readbyte()
            if cmd == SVC_EOF:
                break
            elif cmd == SVC_CONFIGSTRING: 
                self.parse_configstring()
            elif cmd == SVC_BASELINE: 
                self.parse_baseline()
        return self.gamestate

    def parse_configstring(self): 
        i = huffman.readshort()
        string = huffman.readbigstring()
        dest=self.gamestate['configstrings']
        fieldname=str(i)
        output=string
        if CS_STRING_MAP.get(i, None):
            dest=self.gamestate['config']
            fieldname=CS_STRING_MAP.get(i)
        if string.startswith("\\"):
            output={}
            subfields = string.split('\\')
            if not fieldname in dest:
                dest[fieldname]={}
            for x in range(1, len(subfields)-1, 2):
                output[subfields[x]]=subfields[x+1]
        if i >= CS_PLAYERS and i < CS_PLAYERS+MAX_CLIENTS:
            dest=self.gamestate['config']
            fieldname='player'+str(i-CS_PLAYERS)
            subfields = string.split('\\')
            output = {}
            for x in range(0, len(subfields), 2):
                output[subfields[x]]=subfields[x+1]
            output['id']=i-CS_PLAYERS
            if not is_spectator(output):
                self.gamestate['players'].append(output)
            else:
                self.gamestate['spectators'].append(output)
        if i >= CS_SOUNDS and i < CS_SOUNDS+MAX_SOUNDS:
            dest=self.gamestate['config']
            fieldname='sound'+str(i-CS_SOUNDS)
        if i >= CS_LOCATIONS and i < CS_LOCATIONS+MAX_LOCATIONS:
            dest=self.gamestate['config']
            fieldname='location'+str(i-CS_LOCATIONS)
        if i >= CS_TEAM and i < CS_TEAM+MAX_TEAMS:
            dest=self.gamestate['config']
            fieldname='team'+str(i-CS_TEAM+1)
            self.gamestate['teams'].append({'id': i-CS_TEAM+1,
                                            'name': string})
        dest[fieldname]=output

    def parse_baseline(self):
        return {'type': 'baseline'}

    def parse_servercommand(self):
        seq = huffman.readlong()
        string = huffman.readstring()
        
        cmd = string.split()[0]
        string = ' '.join(string.split()[1:])
        if cmd == 'chat':
            player_to=[ x['n'] for x in
                        self.gamestate['players']+self.gamestate['spectators']
                        if x['id']==int(string[2:4])][0]
            player_to=player_to if player_to else 'FIXME'
            return {'type': 'servercommand',
                    'seq': seq,
                    'cmd': cmd,
                    'to': player_to,
                    'string': string}
        return {'type': 'servercommand',
                'seq': seq,
                'cmd': cmd,
                'string': string}

    def parse_playerstate(self):
        return {'type': 'playerstate'}

def is_spectator(userinfo_dict):
    return True if (userinfo_dict['so'] != "0" 
                    or int(userinfo_dict['t']) > 2) else False




#!/bin/env python

## qldemo2json.py
## Shawn Nock, 2014

DEMO_PATH='demos/'
OUTPUT_PATH='api/'

import os
import json

from qldemo import QLDemo

def main():
    for file in os.listdir(DEMO_PATH):
        if not file.endswith('dm_73'):
            continue
        if os.path.isfile(OUTPUT_PATH+''.join([file[:-5],"json"])):
            print("found existing file")
            continue
        print("Processing: {}".format(file))
        d = QLDemo(DEMO_PATH+file)
        with open(OUTPUT_PATH+''.join([file[:-5],"json"]), 'w') as output_file:
            json.dump(list(d), output_file, sort_keys=True, ensure_ascii=False, indent=2)

if __name__ == '__main__':
    main()


